"Python Database Helper"
# import mariadb
import json
from collections.abc import Iterable
from copy import deepcopy
from email.errors import MessageError
from smtplib import SMTPException
from typing import Dict, Tuple, Union

import cx_Oracle
import mysql.connector
import snowflake.connector
from email_sender_simple import send_email as email
from pandas import DataFrame

import pydb.utils as utils

# DB_ROWS = Union[Iterable[], tuple[int, float, str]]
DB_ROWS = Union[Iterable, Tuple[int, float, str]]
DB_INFO = Union[Dict[str, int], Dict[str, str]]
DB_ARGS = Union[int, float, str, Tuple[int, float, str], Iterable]
QUERY_RESULTS = Union[DB_ROWS, None]


class DB_WRAPPER:
    """Base Class for DB Connection Wrapper"""

    def __init__(self, info: DB_INFO, connector, exception=None):
        """DB Connection Wrapper Base Class Constructor
        ----------
        Parameters
        ----------
        info : DB_INFO
            Dictionary containing connection information for the corresponding
            database connector.
            Example:
            ```python
            db_info = {
                "host": "hostname",
                "user": "username",
                "password": "password"
            }
            ```
        connector : database connector functor
            Database connector function
        exception : Exception, optional
            Exception to pass to databsae connector, by default None
        """
        try:
            self._conn = connector(**info)
            self._cur = self._conn.cursor()
            self._Exception = exception if exception is not None else Exception
        except Exception as exception:
            print("Something went wrong connection to DB:", exception)
            raise exception

    def get_connection(self):
        """
        Returns the connection object for the DB

        Returns
        -------
        connection
            Connection object for the DB
        """
        return self._conn

    def get_cursor(self):
        """
        Returns a reference to the cursor object

        Returns
        -------
        cursor
            Cursor object for the DB
        """
        return self._cur

    def get_exception(self):
        """
        Returns the exception handler for the class

        Returns
        -------
        Exception
            Exception handler for the class
        """
        return self._Exception

    def execute(
        self,
        stmt: str,
        args=None,
        dictionary=False,
        dataframe=False,
        dataframe_headers=None,
        logger=None,
        to_csv=None,
    ) -> QUERY_RESULTS:
        """
        Queries the db with <stmt>.  Returns the results.
        Commits the transaction if successful.
        ----------
        Parameters
        ----------
        stmt : str
            Query to execute
        args : int | float | str | List[int, float, str], optional
            Arguments to pass to query, by default None
        dictionary : bool, optional
            Convert query result to dictionary, by default False
        dataframe : bool, optional
            Convert query result to dataframe, by default False
        dataframe_headers : list, optional
            List of column headers for dataframe (can only be used when
            dataframe=True), by default None
        logger : Logger, optional
            Logger object, by default None
        to_csv : str, optional
            Path to save query results as csv, by default None

        Returns
        -------
        QUERY_RESULTS
            Query result
        """
        if dictionary and dataframe:
            raise Exception("Cannot specify both dictionary and dataframe")
        if dictionary and to_csv:
            raise Exception("Cannot specify both dictionary and to_csv")
        try:
            if args is not None:
                # remove any surrounding quotes from the args
                if not isinstance(args, (list, tuple)):
                    args = [args]
                utils.lg(logger, f"Args: {args}", "DEBUG")
                utils.pretty_print_sql(stmt, logger)
                self._cur.execute(stmt, args)
            else:
                utils.lg(logger, "No args specified", "DEBUG")
                self._cur.execute(stmt)
            self._conn.commit()
            utils.lg(logger, f"Finished executing query: {stmt}")
            if to_csv is not None and to_csv != "":
                res: DataFrame = utils.convert_query_result(
                    self._cur, is_dataframe=True, logger=logger
                )
                return utils.results_to_csv(res, to_csv, logger=logger)
            if "SELECT" in stmt:
                return utils.convert_query_result(
                    self._cur, dictionary, dataframe, dataframe_headers, logger
                )
            return None
        except self._Exception as exception:
            utils.lg(
                logger,
                f"Something went wrong: {exception}",
                "ERROR",
            )
            raise exception

    def query(
        self,
        stmt: str,
        args=None,
        dictionary=False,
        dataframe=False,
        dataframe_headers=None,
        logger=None,
        to_csv=None,
    ):
        """
        Queries the db with <stmt>.  Returns the results
        Does not commit the transaction upon completion
        ----------
        Parameters
        ----------
        stmt : str
            Query to execute
        args : int | float | str | List[int, float, str], optional
            Arguments to pass to query, by default None
        dictionary : bool, optional
            Convert query result to dictionary, by default False
        dataframe : bool, optional
            Convert query result to dataframe, by default False
        dataframe_headers : list, optional
            List of column headers for dataframe (can only be used when
            dataframe=True), by default None
        logger : Logger, optional
            Logger object, by default None
        to_csv : str, optional
            Path to save query results as csv, by default None

        Returns
        -------
        QUERY_RESULTS (int | float | str | List[int, float, str])
            Query result
        """
        if dictionary and dataframe:
            raise Exception("Cannot specify both dictionary and dataframe")
        try:
            if args is not None:
                # remove any surrounding quotes from the args
                if not isinstance(args, (list, tuple)):
                    args = [args]
                utils.lg(logger, f"Args: {args}", "DEBUG")
                utils.pretty_print_sql(stmt, logger)
                self._cur.execute(stmt, args)
            else:
                utils.lg(logger, "No args specified", "DEBUG")
                self._cur.execute(stmt)
            utils.lg(logger, f"Finished executing query: {stmt}")
            if to_csv is not None and to_csv != "":
                res: DataFrame = utils.convert_query_result(
                    self._cur, is_dataframe=True, logger=logger
                )
                return utils.results_to_csv(res, to_csv, logger=logger)
            if "SELECT" in stmt:
                return utils.convert_query_result(
                    self._cur, dictionary, dataframe, dataframe_headers, logger
                )
            return None
        except self._Exception as exception:
            utils.lg(logger, f"Something went wrong: {exception}", "ERROR")
            raise exception

    def close(self):
        """Close the db connection and cursor."""
        self._cur.close()
        self._conn.close()

    def execute_procedure(
        self,
        schema_name: str,
        procedure_name: str,
        package_name=None,
        args=None,
        logger=None,
        send_email=False,
        email_info=None,
        error_email_to=None,
    ):
        """
        Executes a procedure with the given arguments or with none if not
        provided
        ----------
        Parameters
        ----------
        schema_name: str
            schema name
        procedure_name: str
            procedure name
        package_name: str, optional
            package name, by default None
        args: str | List[str], optional
            arguments to pass to the procedure, by default None
        logger: Logger, optional
            logger object, by default None
        send_email: bool, optional
            whether or not to send an email, by default False
        email_info: dict, optional
            dictionary containing email information
            email_info = {
                smtp_info: {
                    host: smtp host,
                    port: smtp port,
                },
                email_to: email to send to,
                email_from: email to send from,
                error_email_to: email to send error to,
                subject: email subject,
                message_body: email message body
                attatchments: path to attachment(s)
            }
        """
        stmt = None
        if package_name is None and args is None:
            stmt = f"CALL {schema_name}.{procedure_name}()"
        elif package_name is None and args is not None:
            if isinstance(args, (list, tuple)):
                stmt = f"CALL {schema_name}.{procedure_name}({','.join(args)})"
            else:
                stmt = f"CALL {schema_name}.{procedure_name}({args})"
        elif package_name is not None and args is None:
            stmt = f"CALL {schema_name}.{package_name}.{procedure_name}()"
        elif package_name is not None and args is not None:
            if isinstance(args, (list, tuple)):
                if len(args) > 1:
                    stmt = f"CALL {schema_name}.{package_name}.{procedure_name}({','.join(utils.surround_with_quotes(args))})"
                else:
                    stmt = f"CALL {schema_name}.{package_name}.{procedure_name}({utils.surround_with_quotes(args[0])})"

            else:
                stmt = f"CALL {schema_name}.{package_name}.{procedure_name}({args})"
        if stmt is None:
            raise Exception("No procedure name or args provided")
        try:
            utils.lg(logger, f"Executing procedure: {stmt}")
            self._cur.execute(stmt)
            self._conn.commit()
        except self._Exception as e:
            utils.lg(logger, f"Something went wrong executing the procedure: {e}")
            if email_info is not None:
                temp_email_info = deepcopy(email_info)
                temp_email_info[
                    "message_body"
                ] = f"Something went wrong executing the procedure: {e}"
                temp_email_info[
                    "subject"
                ] = f"Error executing procedure: {procedure_name}"
                if (
                    "error_email_to" in temp_email_info
                    and temp_email_info["error_email_to"] is not None
                ):
                    logger.info(
                        f"Sending error email to: {temp_email_info['error_email_to']}"
                    )
                    temp_email_info["email_to"] = temp_email_info["error_email_to"]
                try:
                    email(
                        **temp_email_info,
                        logger=logger,
                    )
                except (MessageError, SMTPException, TimeoutError) as e:
                    if logger is not None:
                        logger.error(f"Error sending email: {e}")
                    raise e
            raise e
        utils.lg(logger, f"{stmt} executed successfully")

        if send_email and email_info is None:
            utils.lg(logger, "No email info provided")
        elif send_email:
            try:
                if logger is not None:
                    logger.debug(f"Email info: {json.dumps(email_info, indent=4)}")
                email(**email_info, logger=logger)
            except (MessageError, SMTPException, TimeoutError) as e:
                raise e


class MysqlDB(DB_WRAPPER):
    """Mysql Specific Functions"""

    def __init__(self, info):
        """MySQL Connection Wrapper"""
        try:
            DB_WRAPPER.__init__(
                self, info, mysql.connector.connect, mysql.connector.Error
            )
            self._conn = self.get_connection()
            self._cur = self.get_cursor()
        except mysql.connector.Error as exception:
            raise exception

    def get_curdate(self):
        """Returns CURDATE() from MySQL."""
        return self.query("SELECT CURDATE()")

    def get_timestamp(self):
        """Returns CURRENT_TIMESTAMP from MySQL."""
        return self.query("SELECT CURRENT_TIMESTAMP()")

    def table_exists(self, schema: str, table: str):
        stmt = f"""
        SELECT COUNT(*) from information_schema.TABLES
        WHERE TABLE_SCHEMA = '{schema}' and TABLE_NAME = '{table}'
        """
        return self.query(stmt) != 0

    def plsql(
        self,
        in_plsql: str,
        in_name: None,
        logger=None,
        send_email=False,
        email_info=None,
        error_email_to=None,
    ):
        raise NotImplementedError("MySQL does not support PLSQL")


class SnowflakeWrapper(DB_WRAPPER):
    """Snowflake Specific Functions"""

    def __init__(self, info: DB_INFO):
        DB_WRAPPER.__init__(
            self,
            info,
            snowflake.connector.connect,
            snowflake.connector.errors.ProgrammingError,
        )

    def plsql(
        self,
        in_plsql: str,
        in_name: None,
        logger=None,
        send_email=False,
        email_info=None,
        error_email_to=None,
    ):
        raise NotImplementedError("Snowflake does not support PLSQL")


class OracleWrapper(DB_WRAPPER):
    """Oracle specific functions."""

    def __init__(self, info: DB_INFO):
        try:
            DB_WRAPPER.__init__(self, info, cx_Oracle.connect, cx_Oracle.Error)
            self._db_info = info
        except cx_Oracle.Error as e:
            raise e

    def get_incoming_me(self):
        """Returns the ME_INCOMING table from Oracle."""
        return self.query(
            "SELECT POST_DAY_END.PDE_COMMONS_PKG.GET_INCOMING_DAYBREAK_ME_NAME FROM DUAL"
        )

    def get_daybreak_me_target_name(self):
        """Returns the ME_TARGET table from Oracle."""
        return self.query(
            "SELECT POST_DAY_END.PDE_COMMONS_PKG.GET_DAYBREAK_ME_TARGET_NAME FROM DUAL"
        )

    def plsql(
        self,
        in_plsql: str,
        in_name=None,
        logger=None,
        send_email=False,
        email_info=None,
        error_email_to=None,
    ):
        """
        Executes PL/SQL block and optionally sends email.
        ----------
        Parameters
        ----------
        in_plsql : str
            PL/SQL block
        in_name : str, optional
            Name of PL/SQL block
        logger : logging.Logger, optional
            Logger object
        send_email : bool, optional
            Whether or not to send email
        email_info : dict, optional
            Dictionary containing email information
            email_info = {
                smtp_info: {
                    host: smtp host,
                    port: smtp port,
                },
                email_to: email to send to,
                email_from: email to send from,
                error_email_to: email to send error to (defaults to email_to),
                subject: email subject,
                message_body: email message body
                attatchments: path to attachment(s)
            }
        """
        try:
            if not in_plsql:
                raise ValueError("No PL/SQL")
            utils.lg(logger, "Executing PL/SQL block")
            utils.lg(logger, f"{in_plsql}", "DEBUG")
            self._cur.execute(in_plsql)
            self._conn.commit()
        except self._Exception as exception:
            if email_info is not None:
                temp_email_info = deepcopy(email_info)
                temp_email_info[
                    "message_body"
                ] = f"PL/SQL failed to execute. {exception}"
                temp_email_info["subject"] = (
                    f"{in_name} failed to execute"
                    if in_name is not None
                    else "PLSQL failed to execute"
                )
                if (
                    "error_email_to" in temp_email_info
                    and temp_email_info["error_email_to"] is not None
                ):
                    temp_email_info["email_to"] = temp_email_info["error_email_to"]
                    utils.lg(
                        logger, f"Sending error email to {temp_email_info['email_to']}"
                    )
                try:
                    email(
                        **temp_email_info,
                        logger=logger,
                    )
                except (MessageError, SMTPException, TimeoutError) as email_exception:
                    utils.lg(
                        logger,
                        f"Email failed to send: {email_exception}",
                        "ERROR",
                    )
                    raise email_exception
            raise exception
        if send_email and email_info is None:
            utils.lg(logger, "No email info provided")
        elif send_email:
            try:
                utils.lg(logger, f"Email info: {email_info}", "DEBUG")
                if logger is not None:
                    utils.lg(
                        logger,
                        f"Email info: {json.dumps(email_info, indent=4)}",
                        "DEBUG",
                    )
                email(**email_info, logger=logger)
            except (MessageError, SMTPException, TimeoutError) as email_exception:
                utils.lg(logger, f"Email failed to send: {email_exception}", "ERROR")
                raise email_exception
        utils.lg(logger, "plsql executed successfully")
