"""Utility functions for pydb."""
from csv import QUOTE_MINIMAL, QUOTE_NONNUMERIC
from logging import Logger as DefaultLogger
from typing import List, Union

from pandas import DataFrame
from python_logger import Logger
from sqlparse import format

from pydb.db import QUERY_RESULTS
from pydb.managers.database_manager import DatabaseManager


def lg(logger: Union[Logger, DefaultLogger, None], msg: str, level="info"):
    level = level.strip().lower()
    if logger:
        if level == "info":
            logger.info(msg)
        elif level == "debug":
            logger.debug(msg)
        elif level == "warning":
            logger.warning(msg)
        elif level == "error":
            logger.error(msg)
        elif level == "critical":
            logger.critical(msg)
    elif level > "info":
        print(msg)


def pretty_print_sql(sql: str, logger: Union[Logger, DefaultLogger, None] = None):
    """
    Pretty print sql query or PL/SQL block.
    ----------
    Parameters
    ----------
    sql : str
        SQL query or PL/SQL block
    logger : Union[Logger, DefaultLogger, None], optional
        Logger object, by default None

    Returns
    -------
    str
        Pretty printed SQL query or PL/SQL block
    """
    try:
        pretty_query = format(
            sql,
            reindent=True,
            reindent_aligned=False,
            keyword_case="upper",
            identifier_case="upper",
            indent_width=4,
            wrap_after=120,
            truncate_strings=120,
            use_space_around_operators=True,
            # output_format="python",
        )
        lg(logger, pretty_query, "debug")
        return pretty_query
    except Exception as e:
        lg(logger, "Failed to pretty print SQL", "ERROR")
        raise e


def results_to_csv(
    query_results: QUERY_RESULTS,
    out_path: str,
    column_headers=None,
    logger=None,
    quoting=QUOTE_MINIMAL,
    quotechar='"',
    delimiter=",",
    lineterminator="\n",
):
    """
    Execute a query and save the result to a csv file.
    ----------
    Parameters
    ----------
    query_results : QUERY_RESULTS
        Query results
    out_path : str
        Output file path
    column_headers : list, optional
        Explicitly set column headers, by default None
    logger : Union[Logger, DefaultLogger, None], optional
        Logger object, by default None
    quoting : int, optional
        CSV quoting option, by default QUOTE_MINIMAL
    quotechar : str, optional
        CSV quotechar option, by default '\"'
    delimiter : str, optional
        CSV delimiter option, by default ","
    lineterminator : str, optional
        CSV lineterminator option, by default "\\n"
    """
    if not isinstance(query_results, DataFrame):
        lg(logger, "Converting query results to dataframe", "debug")
        try:
            query_results = DataFrame(query_results)
            lg(logger, "Converted query results to dataframe", "debug")
        except Exception as e:
            raise e
        try:
            lg(logger, f"Saving query results to {out_path}", "info")
            query_results.to_csv(
                out_path,
                index=False,
                header=column_headers,
                quoting=quoting,
                quotechar=quotechar,
                sep=delimiter,
                lineterminator=lineterminator,
            )
            lg(logger, f"Wrote query results to {out_path}", "info")
            return out_path
        except Exception as e:
            raise e


def query_to_csv(
    db_info,
    db_type,
    column_headers,
    out_path,
    query,
    args=None,
    logger=None,
    quoting=QUOTE_MINIMAL,
    quotechar='"',
    delimiter=",",
    lineterminator="\n",
):
    """
    Execute a query and save the result to a csv file.
    ----------
    Parameters
    ----------
    db_info : str
        Dictionary of database connection information
        Example:
        ```python
        db_info = {
            "host": "hostname",
            "user": "username",
            "password": "password"
        }
        ```
    db_type : str
        Database type (one of "mysql", "oracle", "snowflake")
    column_headers : list
        List of column headers
    out_path : str
        Output file path
    query : str
        SQL query
    args : list, optional
        List of arguments for query, by default None
    logger : Union[Logger, DefaultLogger, None], optional
        Logger object, by default None
    quoting : int, optional
        CSV quoting option, by default QUOTE_MINIMAL
    quotechar : str, optional
        CSV quotechar option, by default '\"'
    delimiter : str, optional
        CSV delimiter option, by default ","
    lineterminator : str, optional
        CSV lineterminator option, by default "\\n"
    """
    with DatabaseManager(db_info, db_type) as db:
        try:
            df: DataFrame = db.execute(
                query,
                args=args,
                dataframe=True,
                dataframe_headers=column_headers,
                logger=logger,
            )
        except Exception as e:
            lg(logger, f"Failed to execute query: {query}", "error")
            raise e
        try:
            # if not isinstance(header, (list, tuple)):
            #     header = [header]
            # logger.debug("DF:", df)
            # df = df.set_axis(header, axis=1, copy=False)
            # logger.debug("DF:", df)
            df.to_csv(
                out_path,
                index=False,
                header=column_headers,
                quoting=quoting,
                quotechar=quotechar,
                sep=delimiter,
                lineterminator=lineterminator,
            )
        except Exception as e:
            raise e
        return out_path


def convert_query_result(
    cur, is_dictionary=False, is_dataframe=False, dataframe_headers=None, logger=None
) -> QUERY_RESULTS:
    """
    Return the element or List(rows) if multiple results
    Convert from List[List[rows]] -> List[rows] or List[result] -> result

    Parameters
    ----------
    cur : cursor
        Cursor object from database connection
    is_dictionary : bool, optional
        Convert query result to dictionary, by default False
    is_dataframe : bool, optional
        Convert query result to dataframe, by default False
    dataframe_headers : list, optional
        List of column headers for dataframe, by default None
    logger : Logger, optional
        Logger object, by default None

    Returns
    -------
    QUERY_RESULTS
        Query result
    """
    if is_dictionary and is_dataframe:
        raise Exception("Cannot be both dictionary and dataframe")
    try:
        res = cur.fetchall()
    except Exception as e:
        lg(logger, "Failed to fetch results", "ERROR")
        raise e
    if is_dictionary and cur.description is not None:
        return dict_factory(res, cur.description)
    if is_dataframe:
        if len(res) > 0 and dataframe_headers is not None:
            out = DataFrame(res, columns=dataframe_headers)
        elif len(res) > 0:
            out = DataFrame(res)
        else:
            out = None
        return out
    if res is None or (isinstance(res, list) and len(res) == 0):
        return None
    if len(res) == 1:
        if len(res[0]) == 1:
            return res[0][0]
        return res[0]
    return [i[0] if len(i) == 1 else list(i) for i in res]


def surround_with_quotes(s) -> Union[List[str], str]:
    """Surround each element in list with quotes."""
    if isinstance(s, (list, tuple)):
        return [f"'{i}'" for i in s]
    return f"'{s}'"


def dict_factory(rows, description):
    """Converts query result to a list of lists or list of dictionaries."""
    if rows is None:
        return None
    try:
        desc = description
        if desc is None:
            return None
        tdict = [
            dict(zip([col[0].strip().upper() for col in desc], row)) for row in rows
        ]
        return tdict if len(tdict) > 0 else None
    except Exception as e:
        raise e
